@@ -1,31 +1,20 @@
 import ast
+import asyncio
 import json
 import logging
 
 import socketio
 from aiortc import MediaStreamTrack, RTCPeerConnection, RTCIceServer, RTCConfiguration
 from aiortc.contrib.media import MediaRelay, MediaBlackhole
+from aiortc.mediastreams import MediaStreamError
 from aiortc.rtcrtpreceiver import RemoteStreamTrack
+from queue import Queue
+from gspeerconnection.gsmediarecorder import GSMediaRecorder
+from threading import Thread
 
 relay = MediaRelay()
 
 
-class VideoTransformTrack(MediaStreamTrack):
-    kind = "video"
-
-    def __init__(self, track, gsdbs, source, onframe):
-        super().__init__()
-        self.track = track
-        self.onframe = onframe
-        self.source = source
-        self.gsdbs = gsdbs
-
-    async def recv(self):
-        frame = await self.track.recv()
-        self.onframe(self.gsdbs, self.source, frame)
-
-
-
 class GSPeerConnectionWatcher:
 
     @classmethod
@@ -81,14 +70,9 @@ class GSPeerConnectionWatcher:
                         await self.ontrack(self.gsdbs, track, self.target)
                     if self.onframe is not None:
                         self.logger.info("on track received. onframe started")
-                        blackHole = MediaBlackhole()
-                        mediaTrack = VideoTransformTrack(relay.subscribe(track), self.gsdbs, self.target, self.onframe)
-                        blackHole.addTrack(mediaTrack)
-                        await blackHole.start()
-
-                # @track.on("ended")
-                # async def on_ended():
-                #     await self.ontrackended(self.gsdbs, self.target)
+                        gsmediaconsumer = GSMediaConsumer(self.gsdbs, self.target, self.onframe)
+                        gsmediaconsumer.addTrack(track)
+                        await gsmediaconsumer.start()
 
             await self.peerConnections.setLocalDescription(await self.peerConnections.createOffer())
             await self.sio.emit("watcher",
@@ -116,3 +100,69 @@ class GSPeerConnectionWatcher:
         await self.sio.connect(
             f'{connectURL}?gssession={self.gsdbs.cookiejar.get("session")}.{self.gsdbs.cookiejar.get("signature")}{self.gsdbs.credentials["cnode"]}&target={self.target}')
         await self.sio.wait()
+
+
+class FrameBufferThread(Thread):
+    def __init__(self, gsdbs, target, onframe):
+        super(FrameBufferThread, self).__init__()
+        self.gsdbs = gsdbs
+        self.gsdbs.logger.debug("Starting Buffer")
+        self.target = target
+        self.onframe = onframe
+        self.nal_queue = Queue(3)
+
+    def write(self, frame):
+        if not self.nal_queue.full():
+            self.nal_queue.put(frame)
+
+    def run(self):
+        while True:
+            buf = self.nal_queue.get()
+            if buf:
+                self.onframe(self.gsdbs, self.target, buf)
+
+
+class GSMediaConsumer:
+
+    def __init__(self, gsdbs, target, onframe):
+        self.__tracks = {}
+        self.gsdbs = gsdbs
+        self.target = target
+        self.onframe = onframe
+        self.nal_queue = Queue(3)
+        self.broadcastthread = FrameBufferThread(self.gsdbs, self.target, self.onframe)
+        self.broadcastthread.start()
+
+    def addTrack(self, track):
+        if track not in self.__tracks:
+            self.__tracks[track] = None
+
+    async def start(self):
+        """
+        Start discarding media.
+        """
+        for track, task in self.__tracks.items():
+            if task is None:
+                # self.startProcessing(task)
+                # _thread = threading.Thread(target=self.startProcessing,
+                #                            args=(track,))
+                # _thread.start()
+                self.__tracks[track] = asyncio.ensure_future(
+                    self.stream_consume(track, self.gsdbs, self.target, self.onframe))
+
+    async def stream_consume(self, track, gsdbs, target, onframe):
+        while True:
+            try:
+                frame = await track.recv()
+                self.broadcastthread.write(frame)
+            except MediaStreamError:
+                return
+
+    async def stop(self):
+        """
+        Stop discarding media.
+        """
+        for task in self.__tracks.values():
+            if task is not None:
+                task.cancel()
+        self.__tracks = {}
